use std::{collections::HashMap, str::from_utf8, vec};

use actix_web::{get, web, App, HttpResponse, HttpServer};
use argh::FromArgs;
use rand::seq::SliceRandom;
use rust_embed::Embed;
use serde::{Deserialize, Serialize};

static ARTICLE: &str = "<article>";
static GENDER_M: &str = "masculins";
static GENDER_F: &str = "feminins";
static RANDOM: &str = "random";

#[derive(FromArgs)]
#[argh(description = "arguments")]
struct Args {
    #[argh(option, short = 'h', description = "host")]
    host: Option<String>,
    #[argh(option, short = 'p', description = "port")]
    port: Option<u16>,
}

#[derive(Embed)]
#[folder = "dictionaries/"]
struct Dictionaries;

#[derive(Debug, Serialize, Deserialize)]
struct NameGenQuery {
    sequence: Option<String>,
    gender: Option<String>,
    qty: Option<u8>,
    sep: Option<String>,
}

#[get("/")]
async fn index(query: web::Query<NameGenQuery>) -> HttpResponse {
    // Load the dictionaries into memory
    let mut dictionaries = HashMap::new();
    for filename in Dictionaries::iter() {
        if let Some(file) = Dictionaries::get(&filename) {
            let data = from_utf8(file.data.as_ref()).unwrap();
            dictionaries.insert(
                filename.to_string(),
                serde_json::from_str::<Vec<String>>(data).unwrap(),
            );
        }
    }

    let qty = query.0.qty.unwrap_or(10).min(50).max(1);
    let sep = query.0.sep.unwrap_or(" ".to_string());
    let gender = match query.0.gender.as_ref().map(String::as_ref) {
        Some("m") => GENDER_M,
        Some("f") => GENDER_F,
        _ => RANDOM,
    };

    // Limit to 10 parts max
    let q_dicts = query.0.sequence.unwrap_or("prenom patronyme".to_string());
    let mut chosen_dicts = q_dicts.split(" ").collect::<Vec<_>>();
    chosen_dicts.truncate(10);

    let mut names = vec![];

    for _ in 0..qty {
        let mut parts = vec![];

        let gender = if gender == RANDOM {
            vec![GENDER_M, GENDER_F]
                .choose(&mut rand::thread_rng())
                .unwrap()
        } else {
            gender
        };
        println!("{gender}");

        for dict in chosen_dicts.iter() {
            let filename = match *dict {
                "adjectif" | "commun" | "ppasse" | "ppresent" | "prenom" => {
                    Some(format!("{dict}s_{gender}.json"))
                }
                "patronyme" => Some(format!("{dict}s.json")),
                "article" => Some(ARTICLE.to_string()),
                "virgule" => Some(",".to_string()),
                _ => None,
            };
            if let Some(filename) = filename {
                if let Some(data) = dictionaries.get(&filename) {
                    let item = data.choose(&mut rand::thread_rng()).unwrap();
                    parts.push(item.to_string())
                } else {
                    parts.push(filename)
                }
            }
        }

        // Replace the <article> with le/la/l'
        let mut to_replace = vec![];
        for (i, item) in parts.iter().enumerate() {
            if item == ARTICLE {
                // Check the next word
                if let Some(next) = parts.get(i + 1) {
                    if next.starts_with(|c| {
                        vec!['a', 'e', 'i', 'o', 'u', 'é', 'è', 'ê', 'h'].contains(&c)
                    } && !next.starts_with("hy")) {
                        to_replace.push((i, "l'"));
                    } else {
                        to_replace.push((i, (if gender == "masculins" { "le" } else { "la" })))
                    }
                }
            }
        }
        for item in to_replace.iter() {
            parts[item.0] = item.1.to_string();
        }

        let joined = parts
            .join(&sep)
            // Remove space after "l'"
            .replace("l' ", "l'")
            // Remove space before comma
            .replace(" ,", ",");

        names.push(joined);
    }

    HttpResponse::Ok().json(names)
}

#[actix_web::main] // or #[tokio::main]
async fn main() -> std::io::Result<()> {
    let args: Args = argh::from_env();
    let host = args.host.unwrap_or("127.0.0.1".to_string());
    let port = args.port.unwrap_or(8080);

    HttpServer::new(|| App::new().service(index))
        .bind((host, port))?
        .run()
        .await
}
